#ifndef	_stm_h
#	define	_stm_h

#pragma once
#include	<assert.h>
#include	<map>
#include	<set>
#include	<boost/thread.hpp>

using namespace std;
using namespace boost;

class	stmThread;
class	stmCore;

class	genericStmObject	{
public:
	virtual	bool	write_commit(stmThread	*)		=	0;
	virtual	bool	write_rollback(stmThread	*)	=	0;
	virtual	bool	read_commit(stmThread	*)		=	0;
	virtual	bool	read_rollback(stmThread	*)		=	0;
	stmThread	*writeOwner;
	stmCore		*thestmCore;
	mutex		objectLock;
};


class	stmThread {
public:
	mutex		threadLock;
	set<genericStmObject *>  	theReadSet;
	set<genericStmObject *>		theWriteSet;

	int			nestLevel;
	int			thethreadId;
	int			readFaults;
	int			writeFaults;
	int			writeReadCommitFaults;
	int			commits;
	stmThread(int	athreadId) {
		nestLevel = thethreadId = readFaults = writeFaults = writeReadCommitFaults = 0;
		thethreadId = athreadId;
		}
	int			getthreadId()	{	return(thethreadId);	}
};



class	stmRestart {
public:
	enum	{	Abort, Retry, UserRetry	};

	const	int	Exception;
	stmRestart(int	_t) : Exception(_t) {};
	};



#define	begin_transaction	\
	while(1) {									\
		try {									\
			global_stmCore.beginTransaction();

#define	end_transaction	\
	} catch ( stmRestart s ) {										\
		 int r=global_stmCore.handlestmRestart(s.Exception);		\
		 if(r == 0 ){continue;} \
		 if(r == 2 ){break;}	\
		  }															\
	if ( global_stmCore.endTransaction() )							\
		break;														\
	}																


#define	retry_transaction	\
	throw stmRestart(stmRestart::UserRetry);

#define	abort_transaction	\
	throw stmRestart(stmRestart::Abort);

class stmCore
{
public:
	stmCore(void);
	~stmCore(void);

	map<int, stmThread *>	threadMap;
	int			readFaults;
	int			writeFaults;
	int			writeReadCommitFaults;
	int			commits;
	mutex		coreLock;

public:
	int			beginTransaction(char	*debugInformation	=	NULL);
	int			endTransaction();
	int			retryTransaction();
	int			userRetryTransaction();
	int			abortTransaction();
	int			handlestmRestart(int);
	void		printStatistics();
//private:
	static	int	getthreadId();
	stmThread	*getstmThread();
	stmThread	*__getstmThread();
	stmThread	*getstmThread_handleCreate();

	void		ReadFault(genericStmObject	*);
	void		WriteFault(genericStmObject	*);
	stmThread  	*addToReadSet(genericStmObject	*);
	stmThread	*addToWriteSet(genericStmObject	*);
	bool   		__searchForReadConflict(genericStmObject *, stmThread *);
	void		__cleanupFailedTransaction(stmThread *);
};



extern	stmCore	global_stmCore;


template	<class theType>	class	stmObject	:
	public	genericStmObject
{
public:
	theType	theData;
	theType	theCopy;

public:
	stmObject() {
		writeOwner = NULL;
		thestmCore = &global_stmCore;
		}

	stmObject(theType a) {
		writeOwner = NULL;
		thestmCore = &global_stmCore;
		theData = a;
	}

	virtual ~stmObject(void) {
	}

	theType	&	openAndRead() {
		boost::mutex::scoped_lock	lock(objectLock);
		if ( writeOwner ) {
			if ( thestmCore->getthreadId() != writeOwner->getthreadId() )
				thestmCore->ReadFault(this);	// we never return from this function
			else
				return(theCopy);
		}
		thestmCore->addToReadSet(this);
 		return(theData);
	}

	const	theType	&	openAndWrite(const	theType	&aData) {
		boost::mutex::scoped_lock	lock(objectLock);
		if ( writeOwner ) {
			if ( thestmCore->getthreadId() != writeOwner->getthreadId() )
				thestmCore->WriteFault(this);	// we never return from this function
			else {
				theCopy = aData;
				return(theCopy);
			}
		}
		thestmCore->addToReadSet(this);
		writeOwner = thestmCore->addToWriteSet(this);
		// sometimes is useful to read/write TVar's when you aren't in a transaction... but be careful
		if ( ! writeOwner ) {
			theData = aData;
  			return(theData);
			}
		theCopy = aData;
		return(theCopy);
	}

	// useful sometimes for higher-level objects
	theType	&	open() {
		boost::mutex::scoped_lock	lock(objectLock);
		if ( writeOwner ) {
			if ( thestmCore->getthreadId() != writeOwner->getthreadId() )
				thestmCore->WriteFault(this);	// we never return from this function
			else {
				theCopy = theData;
				return(theCopy);
			}
		}
		thestmCore->addToReadSet(this);
		writeOwner = thestmCore->addToWriteSet(this);
		// sometimes is useful to read/write TVar's when you aren't in a transaction... but be careful
		if ( ! writeOwner ) {
			return(theData);
			}
		theCopy = theData;
		return(theCopy);
	}

	operator	theType		()	{ return(openAndRead()); }

	const	theType	&	operator	=	(const	theType	&aData)	{ return(openAndWrite(aData)); }
	const	stmObject<theType>	&	operator	=	(stmObject<theType>	&aData) {
		openAndWrite(aData.openAndRead());
		return(*this);
	}

	theType	*	operator	&	() {
		boost::mutex::scoped_lock	lock(objectLock);
		if ( writeOwner ) {
			if ( thestmCore->getthreadId() != writeOwner->getthreadId() )
				thestmCore->WriteFault(this);	// we never return from this function
			else {
				return(&theCopy);
			}
		}
		thestmCore->addToReadSet(this);
		writeOwner = thestmCore->addToWriteSet(this);
		// sometimes is useful to read/write TVar's when you aren't in a transaction... but be careful
		if ( ! writeOwner ) {   return(&theData);	}
		return(&theCopy);
	}

	bool	write_commit(stmThread	*writingThread) {
		boost::mutex::scoped_lock	lock(objectLock);
		assert(writeOwner != NULL);
		assert(writeOwner == writingThread);
		theData = theCopy;
		writeOwner = NULL;
		return(true);
	}

	bool	write_rollback(stmThread	*writingThread)	{
		boost::mutex::scoped_lock	lock(objectLock);
		assert(writeOwner != NULL);
		writeOwner = NULL;
		return(true);
	}

	bool	read_commit(stmThread	*readingThread) {
		return(true);
		}

	bool	read_rollback(stmThread	*readingThread) {
		return(true);
		}
};


// extends stmObject with some more operators for basic types.

template	<class theType>	class	primitivestmObject	:
	public	stmObject<theType>
{
public:
	const	theType	&	operator	--	() { return(openAndWrite(stmObject<theType>::openAndRead() - 1));}
	const	theType	&	operator	++	() { return(openAndWrite(stmObject<theType>::openAndRead() + 1));}
	const	theType	&	operator	--	(int)	{	theType	&t = stmObject<theType>::open(); openAndWrite(stmObject<theType>::openAndRead() - 1); return(t); }
	const	theType	&	operator	++	(int)	{	theType	&t = stmObject<theType>::open(); openAndWrite(stmObject<theType>::openAndRead() + 1); return(t); }
						operator	theType		()	{ return(stmObject<theType>::operator theType ()); }
	const	theType	&	operator	=	(const	theType	&aData)  {return(stmObject<theType>::operator = (aData));}
	const	stmObject<theType>	&	operator	=	(stmObject<theType>	&aData) { return(stmObject<theType>::operator = (aData)); }
			theType	*	operator	&	()	{return(stmObject<theType>::operator&());}
	};

typedef	primitivestmObject<char>			tm_char;
typedef	primitivestmObject<unsigned char>	tm_uchar;
typedef	primitivestmObject<unsigned char>	tm_byte;
typedef	primitivestmObject<int>				tm_int;
typedef	primitivestmObject<unsigned int>	tm_uint;
typedef	primitivestmObject<short>			tm_short;
typedef	primitivestmObject<unsigned short>	tm_ushort;
typedef	primitivestmObject<long>			tm_long;
typedef	primitivestmObject<unsigned long>	tm_ulong;
typedef	primitivestmObject<float>			tm_float;
typedef	primitivestmObject<double>			tm_double;

extern	stmCore	global_stmCore;

#endif

